require 'open-uri'
require 'rubygems'
require 'mechanize'
require 'hpricot'

module Scrubyt
##
#=<tt>Driving the whole extraction process</tt>
#
#Extractor is a performer class - it gets an extractor definition and carries
#out the actions and evaluates the wrappers sequentially. 
#
#Originally also the navigation actions were here, but since the class got too
#big, they were factored out to an own class, NavigationAction.
  class Extractor 
    #The definition of the extractor is passed through this method
    def self.define(mode=nil, &extractor_definition)
      @@mode = mode
      #We are keeping the relations between the detail patterns and their root patterns
      @@detail_extractor_to_pattern_name = {}
      @@detail_pattern_relations = {}
      #root pattern -> URIBuilder mapping
      @@next_patterns = {}
      mode_name = (mode == :production ? 'Production' : 'Learning')
      puts "[MODE] #{mode_name}"
      NavigationActions.new
      @@evaluation_context = EvaluationContext.new
      #Hack up an artificial root pattern (i.e. do not return the pattern which 
      #is the root one in the user's definition, but rather the real (invisible)
      #root pattern
      evaluated_extractor = (class_eval(&extractor_definition))
      if evaluated_extractor == nil
        puts "No extractor defined, exiting..."
        exit
      end
      root_pattern = evaluated_extractor.parent
      #Recursively match data based on examples
      @@evaluation_context.setup_examples
      #Once all is set up, evaluate the extractor from the root pattern!
      evaluate_extractor(root_pattern)
      #Apply all postprocess steps
      PostProcessor.apply_post_processing(root_pattern)
      #Return the root pattern
      puts "Extraction finished succesfully!"
      root_pattern
    end

  #Evaluate a subexttractor (i.e. an extractor on a detail page).
  #The url passed to this function is automatically loaded.
  #The definition of the subextractor is passed as a block
  #
  #!!!! THIS CODE IS A MESS, IT needs to be refactored ASAP....
  def self.evaluate_subextractor(url, parent_pattern)
    if @@detail_pattern_relations.keys.include? @@detail_extractor_to_pattern_name[parent_pattern.referenced_extractor]
      detail_root = @@detail_pattern_relations[@@detail_extractor_to_pattern_name[parent_pattern.referenced_extractor]].parent
      detail_root.result = Result.new
      detail_root.last_result = nil
      @@original_evaluation_context.push @@evaluation_context
      @@evaluation_context = EvaluationContext.new
      @@evaluation_context.clear_sources_and_sinks detail_root
      FetchAction.restore_host_name
      fetch url
      @@evaluation_context.extractor = self
      @@evaluation_context.root_pattern = detail_root      
      @@evaluation_context.attach_current_document     
      evaluate_extractor detail_root      
      @@evaluation_context = @@original_evaluation_context.pop
      detail_root.to_xml
    else      
      @@original_evaluation_context ||= []
      FetchAction.restore_host_name
      @@original_evaluation_context.push @@evaluation_context
      @@evaluation_context = EvaluationContext.new
      fetch url
      evaluated_extractor = (class_eval(&parent_pattern.referenced_extractor))
      root_pattern = evaluated_extractor.parent
      @@detail_pattern_relations[@@detail_extractor_to_pattern_name[parent_pattern.referenced_extractor]] = root_pattern.children[0]
      @@evaluation_context.setup_examples
      evaluate_extractor(root_pattern)
      #Apply all postprocess steps
      PostProcessor.apply_post_processing(root_pattern)
      #Return the root pattern
      #puts "Extracted detail page"
      @@evaluation_context = @@original_evaluation_context.pop
      root_pattern.to_xml
    end
  end

  #build the current wrapper    
  def self.method_missing(method_name, *args, &block)
    if NavigationActions::KEYWORDS.include? method_name.to_s
      NavigationActions.send(method_name, *args)
      return
    end
    pattern = Scrubyt::Pattern.new(method_name.to_s, *args)
    check_if_shortcut_pattern(pattern)
    check_if_detail_page(pattern, args)
    pattern.evaluation_context = @@evaluation_context
    if @parent == nil
      if method_name.to_s == 'next_page'
        @@evaluation_context.setup_uri_builder(pattern, args)
        @@next_patterns[@@last_root_pattern] = @@evaluation_context.uri_builder
        p @@last_root_pattern.children[0].name
        return @@last_pattern
      else
        #Create a root pattern
        root_pattern = Scrubyt::Pattern.new('root', :type => :root) 
        @@last_root_pattern = root_pattern
        root_pattern.evaluation_context = @@evaluation_context
        @@evaluation_context.root_pattern = root_pattern
        @@evaluation_context.extractor = self
        #add the currently active document to the root pattern 
        @@evaluation_context.attach_current_document
        @@evaluation_context.root_pattern.add_child_pattern(pattern)
        @@evaluation_context.block_count = 0
      end
    else
      @parent.add_child_pattern(pattern) if @parent != nil
    end
    if block_given?
      @@evaluation_context.block_count = @@evaluation_context.block_count + 1
      @stack ||=[]
      @parent = pattern
      @stack.push @parent
      class_eval(&block)
      @stack.pop
      @parent = @stack.last
    end    
    @@last_pattern = pattern
  end
  
  #Shortcut patterns, as their name says, are a shortcut for creating patterns
  #from predefined rules; for example:
  #
  #  detail_url 
  #  
  #  is equivalent to
  #  
  #  detail_url 'href', type => :attribute
  #  
  #i.e. the system figures out on it's own that because of the postfix, the
  #example should be looked up (but it should never override the user input!)
  #another example (will be available later):
  #
  # every_img
  # 
  # is equivivalent to
  # 
  # every_img '//img'
  # 
  def self.check_if_shortcut_pattern(pattern)
    case pattern.name
      when /.+_url/
        #make sure that we are not overriding the user's settings
        if !pattern.examples
          pattern.filters[0].example = 'href'
          pattern.type = Scrubyt::Pattern::PATTERN_TYPE_ATTRIBUTE
        end
    end
  end
  
  #Check whether the currently created pattern is a detail pattern (i.e. it refrences
  #a subextractor). Also check if the currently created pattern is 
  #an ancestor of a detail pattern , and store this in a hash if yes (to be able to 
  #traverse the pattern structure on detail pages as well).
  def self.check_if_detail_page(pattern, args)
    return if args.size == 0
    return if !args[0].is_a? Hash 
    return if !args[0][:references]
    referenced_extractor = args[0][:references]
    pattern.type = Scrubyt::Pattern::PATTERN_TYPE_DETAIL
    pattern.referenced_extractor = referenced_extractor
    @@detail_extractor_to_pattern_name[referenced_extractor] ||= []
    @@detail_extractor_to_pattern_name[referenced_extractor] = @@detail_extractor_to_pattern_name[referenced_extractor] << pattern    
  end
    
  def self.get_hpricot_doc
    NavigationActions.get_hpricot_doc
  end
  
  def self.get_current_doc_url
    NavigationActions.get_current_doc_url
  end
  
  def self.get_detail_pattern_relations
    @@detail_pattern_relations
  end
  
  def self.get_mode
    @@mode
  end
  
private
    def self.evaluate_extractor(root_pattern)
      if @@next_patterns[root_pattern]
        current_page_count = 1
        loop do
          really_evaluate_extractor(root_pattern)
          break if (@@next_patterns[root_pattern].limit == current_page_count || @@evaluation_context.crawl_to_new_page(root_pattern, @@next_patterns[root_pattern]) == nil)
          current_page_count += 1 if @@next_patterns[root_pattern].limit != nil
        end
      else 
        really_evaluate_extractor(root_pattern)
      end
    end  
    
    def self.really_evaluate_extractor(pattern)
      pattern.evaluate
      pattern.children.each { |child| really_evaluate_extractor child } 
    end #end of method evaluate_wrapper    
  end #end of class Extractor
end #end of module Scrubyt